/**
 *	hecomsub  --  communication subroutines
 *		Kazuro Furukawa, KEK, Japan (furukawa@kek.jp)
 *		(if you modified this code, please inform furukawa)
 *
 *	K.Furukawa  Mar.24.1992.(2)  comsub
 *	K.F         Aug.29.1988.     swmodule for os9
 *	K.F         Jul.8.1993.      hecomsub
 *	Modifications
 *	K.F         Sep.15.1993.     no warning message in he_open
 *	K.F         Sep.15.1993.     accept multiple clients
 *	K.F         Sep.15.1993.     fix bug in semaphore handling
 *	K.F         Sep.18.1993.     change header size
 *
 *	This module defines communication routines using semaphore
 *	and shared memory.
 *	 int he_open( portnum, buffersize, opt )
 *	 int he_send( buffer, size, opt )
 *	 int he_accept( buffer, size, opt )
 *	 int he_reply( buffer, size, opt )
 *	 int he_receive( buffer, size, opt )
 *	 int he_close( opt )
 *	 int he_init( opt )
 *	If option is 1, error message is printed on to stderr 
 *	when error occured.
 *	If option is 3, error and information message is printed on to stderr 
 *	when error occured.
 *	Error codes errno returned from the system subroutines are preserved 
 *	on return to the caller.
 *	It uses one shared memory block and three semaphores internally.
 */

static char hecomsub_what[] = 
"@(#)hecomsub.c v1.4.1 communication subroutines, K.Furukawa, Sep.1993"; /*
**/

#include "shmsub.h"			/* shmsub definitions */
#include "semsub.h"			/* semsub definitions */
#include <stdio.h>

#define HEAD 8				/* header size */

void he_errmsg();

extern int cm_errno;

static int he_port;			/* port number */
static int he_size;			/* shared memory size */
static void * he_sup;			/* shared memory pointer */
static void * he_sud;			/* shared memory descriptor */
static void * he_sdp;			/* shared memory pointer */
static void * he_sdd;			/* shared memory descriptor */
static int he_lus;			/* semaphore for up locking */
static int he_lds;			/* semaphore for down locking */
static int he_ups;			/* semaphore for up-line */
static int he_dns;			/* semaphore for down-line */
#define SIZEUP *(((int *)he_sup)+1)	/* up-line size */
#define SIZEDN *(((int *)he_sdp)+1)	/* down-line size */
#define BUFFUP (((char *)he_sup)+HEAD)	/* up-line buffer pointer */
#define BUFFDN (((char *)he_sdp)+HEAD)	/* down-line buffer pointer */

/**
 *  he_open
 *	he_open prepares communication
 *
 *    int he_open( portnum, buffersize, option )
 *
 *	portnum is the communication port number, which is used as
 *	shared memory and semaphore key numbers
 *	buffersize is the maximum communication buffer size
 *	if option is 1 error message is sent to stderr
 *	if option is 3 error/information message is sent to stderr
 *
 *    return:	positive/0	success
 *		<0		failure
 */
int 
 he_open( portnum, size, option )
int portnum;			/* port number */
int size;			/* communication buffer size */
int option;			/* option 1: error, 2: info. message */
/*
**/
{
    int stat;
    long info[4];

    he_port = portnum;

    /* create shared memory */
    stat = cm_create( portnum, size+HEAD, CM_PALL, 0);
    if ((stat != CM_EXIST) && (stat != CM_SUCCESS))  {
	if (option)  he_errmsg("cm_create failure");
	return stat;
    }
    stat = cm_create( portnum+1, size+HEAD, CM_PALL, 0);
    if ((stat != CM_EXIST) && (stat != CM_SUCCESS))  {
	if (option)  he_errmsg("cm_create failure");
	return stat;
    }

    /* map shared memory */
    he_sup = cm_map(portnum, option);
    if (he_sup == NULL)  {
	if (option)  he_errmsg("cm_map failure");
	return cm_errno;
    }
    he_sud = cm_desc;
    he_sdp = cm_map(portnum+1, option);
    if (he_sdp == NULL)  {
	if (option)  he_errmsg("cm_map failure");
	return cm_errno;
    }
    he_sdd = cm_desc;

    /* get shared memory info */
    stat = cm_info(portnum, info, option);
    if (stat)  {
	if (option)  he_errmsg("cm_info failure");
	return stat;
    }
    he_size = info[0];
    stat = cm_info(portnum+1, info, option);
    if (stat)  {
	if (option)  he_errmsg("cm_info failure");
	return stat;
    }
    if (he_size < info[0])  he_size = info[0];

    /* create semaphore */
    stat = fs_create( portnum, FS_PALL, 0);
    if ((stat != FS_EXIST) && (stat != FS_SUCCESS))  {
	if (option)  he_errmsg("fs_create failure");
	return stat;
    }
    stat = fs_create( portnum+1, FS_PALL, 0);
    if ((stat != FS_EXIST) && (stat != FS_SUCCESS))  {
	if (option)  he_errmsg("fs_create failure");
	return stat;
    }
    stat = fs_create( portnum+2, FS_PALL, 0);
    if ((stat != FS_EXIST) && (stat != FS_SUCCESS))  {
	if (option)  he_errmsg("fs_create failure");
	return stat;
    }
    stat = fs_create( portnum+3, FS_PALL, 0);
    if ((stat != FS_EXIST) && (stat != FS_SUCCESS))  {
	if (option)  he_errmsg("fs_create failure");
	return stat;
    }
    
    /* attach semaphore */
    he_lus = fs_attach( portnum, option);
    if (he_lus < 0)  {
	if (option)  he_errmsg("fs_attach failure");
	return he_lus;
    }
    he_lds = fs_attach( portnum+1, option);
    if (he_lds < 0)  {
	if (option)  he_errmsg("fs_attach failure");
	return he_lds;
    }
    he_ups = fs_attach( portnum+2, option);
    if (he_ups < 0)  {
	if (option)  he_errmsg("fs_attach failure");
	return he_ups;
    }
    he_dns = fs_attach( portnum+3, option);
    if (he_dns < 0)  {
	if (option)  he_errmsg("fs_attach failure");
	return he_dns;
    }

#ifdef DEBUG
    he_fsdump("he_open");
#endif
    return 0;
}

/**
 *  he_init
 *	he_init initialize communication internals
 *
 *    int he_init( option )
 *
 *	if option is 1 error message is sent to stderr
 *	if option is 3 error/information message is sent to stderr
 *
 *    return:	positive/0	success
 *		<0		failure
 */
int 
 he_init( option )
int option;			/* option 1: error, 2: info. message */
/*
**/
{
    int stat;

    /* initialize semaphore */
    stat = fs_reset( he_lus, option);
    if (stat)  {
	if (option)  he_errmsg("fs_reset failure");
	return stat;
    }
    stat = fs_reset( he_lds, option);
    if (stat)  {
	if (option)  he_errmsg("fs_reset failure");
	return stat;
    }
    stat = fs_reset( he_ups, option);
    if (stat)  {
	if (option)  he_errmsg("fs_reset failure");
	return stat;
    }
    stat = fs_reset( he_dns, option);
    if (stat)  {
	if (option)  he_errmsg("fs_reset failure");
	return stat;
    }

    /* release resources */
    stat = fs_release(he_lus, option);
    if (stat)  {
	if (option)  he_errmsg("fs_release failure");
	return stat;
    }
    stat = fs_release(he_lds, option);
    if (stat)  {
	if (option)  he_errmsg("fs_release failure");
	return stat;
    }

#ifdef DEBUG
    he_fsdump("he_init");
#endif
    return 0;
}

/**
 *  he_send
 *	he_send sends message to server
 *
 *    int he_send( buffer, size, option )
 *
 *	if option is 1 error message is sent to stderr
 *	if option is 3 error/information message is sent to stderr
 *
 *    return:	positive/0	success, sent buffer size
 *		<0		failure
 */
int 
 he_send( buffer, size, option )
void * buffer;			/* message buffer */
int size;			/* communication buffer size */
int option;			/* option 1: error, 2: info. message */
/*
**/
{
    int stat;

    /* wait line locking ready */
    stat = fs_wait(he_lus, option);
    if (stat)  {
	if (option)  he_errmsg("fs_wait failure");
	return stat;
    }

    if (size > he_size)  size = he_size;
    SIZEUP = size;
    memcpy(BUFFUP, buffer, size);

    /* release up-line */
    stat = fs_release(he_ups, option);
    if (stat)  {
	if (option)  he_errmsg("fs_release failure");
	return stat;
    }

#ifdef DEBUG
    he_fsdump("he_send");
#endif
    return size;
}

/**
 *  he_accept
 *	he_accept receives message from client
 *
 *    int he_accept( buffer, size, option )
 *
 *	if option is 1 error message is sent to stderr
 *	if option is 3 error/information message is sent to stderr
 *
 *    return:	positive/0	success, received buffer size
 *		<0		failure
 */
int 
 he_accept( buffer, size, option )
void * buffer;			/* message buffer */
int size;			/* communication buffer size */
int option;			/* option 1: error, 2: info. message */
/*
**/
{
    int stat;

    /* wait up-line ready */
    stat = fs_wait(he_ups, option);
    if (stat)  {
	if (option)  he_errmsg("fs_wait failure");
	return stat;
    }

    if (size > SIZEUP)  size = SIZEUP;
    memcpy(buffer, BUFFUP, size);

    /* release locking */
    stat = fs_release(he_lus, option);
    if (stat)  {
	if (option)  he_errmsg("fs_release failure");
	return stat;
    }

#ifdef DEBUG
    he_fsdump("he_accept");
#endif
    return size;
}

/**
 *  he_reply
 *	he_reply sends back message to client
 *
 *    int he_reply( buffer, size, option )
 *
 *	if option is 1 error message is sent to stderr
 *	if option is 3 error/information message is sent to stderr
 *
 *    return:	positive/0	success, sent buffer size
 *		<0		failure
 */
int 
 he_reply( buffer, size, option )
void * buffer;			/* message buffer */
int size;			/* communication buffer size */
int option;			/* option 1: error, 2: info. message */
/*
**/
{
    int stat;

    /* wait line locking ready */
    stat = fs_wait(he_lds, option);
    if (stat)  {
	if (option)  he_errmsg("fs_wait failure");
	return stat;
    }

    if (size > he_size)  size = he_size;
    SIZEDN = size;
    memcpy(BUFFDN, buffer, size);

    /* release down-line */
    stat = fs_release(he_dns, option);
    if (stat)  {
	if (option)  he_errmsg("fs_release failure");
	return stat;
    }

#ifdef DEBUG
    he_fsdump("he_reply");
#endif
    return size;
}

/**
 *  he_receive
 *	he_receive receives message from server
 *
 *    int he_receive( buffer, size, option )
 *
 *	if option is 1 error message is sent to stderr
 *	if option is 3 error/information message is sent to stderr
 *
 *    return:	positive/0	success, received buffer size
 *		<0		failure
 */
int 
 he_receive( buffer, size, option )
void * buffer;			/* message buffer */
int size;			/* communication buffer size */
int option;			/* option 1: error, 2: info. message */
/*
**/
{
    int stat;

    /* wait down-line ready */
    stat = fs_wait(he_dns, option);
    if (stat)  {
	if (option)  he_errmsg("fs_wait failure");
	return stat;
    }

    if (size > SIZEDN)  size = SIZEDN;
    memcpy(buffer, BUFFDN, size);

    /* release locking */
    stat = fs_release(he_lds, option);
    if (stat)  {
	if (option)  he_errmsg("fs_release failure");
	return stat;
    }

#ifdef DEBUG
    he_fsdump("he_receive");
#endif
    return size;
}

/**
 *  he_close
 *	he_close terminates communication
 *	this version does not remove shared memory and semaphore
 *
 *    int he_close( option )
 *
 *	if option is 1 error message is sent to stderr
 *	if option is 3 error/information message is sent to stderr
 *
 *    return:	positive/0	success
 *		<0		failure
 */
int 
 he_close( option )
int option;			/* option 1: error, 2: info. message */
/*
**/
{
    int stat;

    /* unmap shared memory */
    stat = cm_unmap(he_sud, option);
    if (stat)  {
	if (option)  he_errmsg("cm_unmap failure");
	return stat;
    }
    stat = cm_unmap(he_sdd, option);
    if (stat)  {
	if (option)  he_errmsg("cm_unmap failure");
	return stat;
    }

    /* detach semaphore */
    stat = fs_detach(he_port, option);
    if (stat)  {
	if (option)  he_errmsg("fs_detach failure");
	return stat;
    }
    stat = fs_detach(he_port+1, option);
    if (stat)  {
	if (option)  he_errmsg("fs_detach failure");
	return stat;
    }
    stat = fs_detach(he_port+2, option);
    if (stat)  {
	if (option)  he_errmsg("fs_detach failure");
	return stat;
    }
    stat = fs_detach(he_port+3, option);
    if (stat)  {
	if (option)  he_errmsg("fs_detach failure");
	return stat;
    }

    return 0;
}

/*
 *  he_errmsg
 *	print error message
 */
void
 he_errmsg( /* flg, nm, */ msg )
/* int flg;				/* flag to use perror */
/* char * nm;				/* routine name where error occur*/
char * msg;				/* error message to display */
{
    extern int errno;
    int errtmp;
    errtmp = errno;
/*     if ( flg )  perror( nm ); */
/*     fprintf( stderr, "%s: %s\n", nm, msg ); */
    fprintf( stderr, "%s\n", msg );
    errno = errtmp;
}

#ifdef DEBUG
/*
 *  he_fsdump
 *	dump semaphores
 */
static int
 he_fsdump( msg )
{
    long info[4];

    fprintf( stderr, "%s\n", msg );
    fs_info( he_port, info, 1 );
    fprintf( stderr, "up-lock:   id=%x value=%ld\n", he_lus, info[0] );
    fs_info( he_port+1, info, 1 );
    fprintf( stderr, "down-lock: id=%x value=%ld\n", he_lds, info[0] );
    fs_info( he_port+2, info, 1 );
    fprintf( stderr, "up-sync:   id=%x value=%ld\n", he_ups, info[0] );
    fs_info( he_port+3, info, 1 );
    fprintf( stderr, "down-sync: id=%x value=%ld\n", he_dns, info[0] );
}
#endif
